/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package dialogs;

import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.ShellAdapter;
import org.eclipse.swt.events.ShellEvent;
import org.eclipse.swt.graphics.Cursor;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.PaletteData;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Shell;

import utils.ImageUtils;

/**
 * Instances of this class create a full screen preview
 * of the given image. The image is resized to make a
 * perfect match with the screen size
 */
public class FullScreenPreview {

	/**
	 * This dialog's Shell parent
	 */
	private final Shell parent;

	/**
	 * The shell for the window
	 */
	private Shell sShell;

	/**
	 * key listener used to notify the shell that a key was pressed
	 */
	private Listener sShellKeyListener;

	/**
	 * mouse listener used to notify the shell that a mouse button was pressed
	 */
	private Listener mouseListener;

	/**
	 * The composite used to draw the image
	 */
	private Composite composite;

	/**
	 * The given image
	 */
	private BufferedImage image;

	/**
	 * The cursor used when mouse is over the window
	 */
	private Cursor cursor;

	/**
	 * Creates a new instance of this class given
	 * it's Shell parent and the desired image
	 * @param parent the Shell parent
	 * @param image the desired image
	 */
	public FullScreenPreview(Shell parent, BufferedImage image) {
		this.parent = parent;
		this.image = image;
		createSShell();
	}

	/**
	 * This method initializes sShell
	 */
	private void createSShell() {

		//grid layout for the shell
		GridLayout gridLayoutSShell = new GridLayout();
		gridLayoutSShell.marginHeight = 0;
		gridLayoutSShell.marginWidth = 0;

		//initialize the shell
		sShell = new Shell(parent, SWT.NO_TRIM | SWT.ON_TOP);
		sShell.setBounds(0, 0, Display.getCurrent().getBounds().width, Display.getCurrent().getBounds().height);
		sShell.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_BLACK));
		sShell.setLayout(gridLayoutSShell);

		// create a cursor with a transparent image
		org.eclipse.swt.graphics.Color white = Display.getCurrent().getSystemColor(SWT.COLOR_WHITE);
		org.eclipse.swt.graphics.Color black = Display.getCurrent().getSystemColor(SWT.COLOR_BLACK);
		PaletteData palette = new PaletteData(new RGB[] { white.getRGB(), black.getRGB() });
		ImageData sourceData = new ImageData(16, 16, 1, palette);
		sourceData.transparentPixel = 0;
		cursor = new Cursor(Display.getCurrent(), sourceData, 0, 0);
		sShell.setCursor(cursor);

		//sometimes the sShellKeyListener is not called because sShell has no focus but is active. this will fix it
		sShell.addShellListener(new ShellAdapter() {
			public void shellActivated(ShellEvent e) {
				sShell.setFocus();
				super.shellActivated(e);
			}
		});

		//notify the shell that keyboard keys were pressed
		sShellKeyListener = new Listener() {
			public void handleEvent(Event event) {
				//certify that sShell is the active shell and the ESC or F11 was pressed
				if (sShell.getDisplay().getActiveShell().equals(sShell)) {
					switch (event.keyCode) {
					case SWT.ESC:
						sShell.close();
						break;
					case SWT.F11:
						sShell.close();
						break;

					default:
						break;
					}
				}
			}
		};

		//mouse listener to close the window on mouse click
		mouseListener = new Listener() {
			public void handleEvent(Event event) {
				//certify that sShell is the active shell and the mouse button was pressed
				if (sShell.getDisplay().getActiveShell().equals(sShell)) {
					sShell.close();
				}
			}

		};

		//add the shell keyboard and mouse listeners
		sShell.getDisplay().addFilter(SWT.MouseUp, mouseListener);
		sShell.getDisplay().addFilter(SWT.KeyDown, sShellKeyListener);

		//dispose the listeners when shell is disposed
		sShell.addDisposeListener(new DisposeListener() {
			public void widgetDisposed(DisposeEvent e) {
				sShell.getDisplay().removeFilter(SWT.KeyDown, sShellKeyListener);
				sShell.getDisplay().removeFilter(SWT.MouseUp, mouseListener);
				cursor.dispose();
			}
		});

		//create the composite that holds the image and open the window
		createComposite();
		sShell.open();

		//create a better (slower) resizing of the displayed image and update it
		Display.getCurrent().asyncExec(new Runnable() {
			public void run() {
				image = resizeImage(image, Display.getCurrent().getBounds().width, Display.getCurrent().getBounds().height, AffineTransformOp.TYPE_BICUBIC);
				if (!composite.isDisposed()) {
					composite.getBackgroundImage().dispose();
					composite.setBackgroundImage(new Image(Display.getCurrent(), ImageUtils.convertToSWT(image)));
				}
				image = null;
			}
		});
	}

	/**
	 * This method initializes composite
	 */
	private void createComposite() {

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.CENTER;
		gridDataComposite.grabExcessHorizontalSpace = true;

		BufferedImage resizedImage = resizeImage(image, Display.getCurrent().getBounds().width, Display.getCurrent().getBounds().height, AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
		Image swtImage = new Image(Display.getCurrent(), ImageUtils.convertToSWT(resizedImage));
		resizedImage = null;

		//initialize the composite
		composite = new Composite(sShell, SWT.NONE);
		composite.setLayout(new GridLayout());
		composite.setLayoutData(gridDataComposite);
		composite.setBackgroundImage(swtImage);

		//adjust the composite size to the given image size
		gridDataComposite.widthHint = composite.getBackgroundImage().getBounds().width;
		gridDataComposite.heightHint = composite.getBackgroundImage().getBounds().height;

		//dispose the create SWT image when composite is disposed
		composite.addDisposeListener(new DisposeListener() {
			public void widgetDisposed(DisposeEvent e) {
				composite.getBackgroundImage().dispose();
			}
		});

	}

	/**
	 * Resize the given image to the desired width and height,
	 * using the resize quality which could be one of the
	 * following algorithms.
	 * @see AffineTransformOp#TYPE_NEAREST_NEIGHBOR
	 * @see AffineTransformOp#TYPE_BICUBIC
	 * 
	 * @param image the image to be resized
	 * @param width the desired image width
	 * @param height the desired image height
	 * @param resizeQuality the resizing quality to be used on resizing
	 * @return a new resized image
	 */
	private BufferedImage resizeImage(BufferedImage image, int width, int height, int resizeQuality) {

		double ratio = ImageUtils.getResizingRatio(image, width, height);

		AffineTransform tx = new AffineTransform();
		tx.scale(ratio, ratio);
		AffineTransformOp op = null;

		switch (resizeQuality) {
		case AffineTransformOp.TYPE_NEAREST_NEIGHBOR:
			op =  new AffineTransformOp(tx, AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
			break;
		case AffineTransformOp.TYPE_BICUBIC:
			op =  new AffineTransformOp(tx, AffineTransformOp.TYPE_BICUBIC);
			break;

		default:
			return null;
		}

		//create the resized image
		return op.filter(image, null);
	}

	/**
	 * Returns the shell used by this window
	 * @return the shell used by this window
	 */
	public Shell getSShell() {
		return sShell;
	}
}